-- $Id: Evaluator.hs 74 2008-01-05 18:37:58Z inf6254 $
-- -----------------------------------------------------------------------------
module Evaluator where

-- -----------------------------------------------------------------------------
-- Standard Evaluierer
-- -----------------------------------------------------------------------------
data Term               =  Con Int 
                        |  Div Term Term
                           deriving Show

answer, wrong           :: Term
answer                  =  Div (Div (Con 1972) (Con 2)) (Con 23)
wrong                   =  Div (Con 2) (Div (Con 1) (Con 0))

eval                    :: Term -> Int
eval (Con x)            =  x
eval (Div t u)          =  (eval t) `div` (eval u)

-- -----------------------------------------------------------------------------
-- Exception Evaluierer
-- -----------------------------------------------------------------------------
data Exc a              =  Raise Exception
                        |  Return a
type Exception          =  String

instance Show a         => Show (Exc a) where
    show (Raise e)      =  "exception: " ++ e
    show (Return x)     =  "value: " ++ show x

evalExc                 :: Term -> Exc Int
evalExc (Con x)         =  Return x
evalExc (Div t u)       =  h (evalExc t)
                           where
                           h (Raise e)  = Raise e
                           h (Return x) = h' (evalExc u)
                             where
                             h' (Raise e') = Raise e'
                             h' (Return y) 
                               | y == 0 = Raise "division by zero"
                               | y /= 0 = Return (x `div` y)

-- -----------------------------------------------------------------------------
-- Trace Evluierer
-- -----------------------------------------------------------------------------
newtype Out a           =  MkOut (Output,a)
type Output             =  String

instance Show a         => Show (Out a) where
    show (MkOut (ox,x)) =  ox ++ "value: " ++ show x

line                    :: Term -> Int -> Output
line t x                =  "term: " ++ show t ++ 
                           ", yields: " ++ show x ++ "\n"

evalOut                 :: Term -> Out Int
evalOut (Con x)         =  MkOut (line (Con x) x,x)
evalOut (Div t u)       =  MkOut (ox ++ oy ++ line (Div t u) z,z)
                           where
                           MkOut (ox,x) = evalOut t
                           MkOut (oy,y) = evalOut u
                           z            = x `div` y

-- -----------------------------------------------------------------------------
-- State Evaluierer
-- -----------------------------------------------------------------------------
newtype St a            =  MkSt (State -> (a,State))
type State              =  Int

apply                   :: St a -> State -> (a,State)
apply (MkSt f) s        =  f s

instance Show a         => Show (St a) where
    show f              =  "value: " ++ show x ++
                           ", count: " ++ show s
                           where (x,s) = apply f 0

evalSt                  :: Term -> St Int
evalSt (Con x)          =  MkSt f 
                           where f s = (x,s)
evalSt (Div t u)        =  MkSt f
                           where
                           f s = (x `div` y, s'' + 1)
                                 where
                                 (x,s')  = apply (evalSt t) s
                                 (y,s'') = apply (evalSt u) s'